#include "mbed.h"
#include "main.h"
#include "SWO.h"
#include "pca9685.h"
#include "tusb322.h"
#include "ds3231.h"
#include "ioc.h"
#include "animation.h"

I2C i2c(PA_10, PA_9);
// SWO_Channel swo("swo");

bool RtcTick, RefreshTick, RngTick;

typedef enum {
    Decrementing = 0,
    Incrementing = 1,
} DigitState;

typedef struct {
    DigitState CurrentState;
    int Value;
    bool Updated;
} Digit;

typedef struct {
    Digit Digits[10];
    bool Refresh;
    int FadeDuration;
} Tube;

Tube Tubes[4];  // Active per-tube configuration
Digit Dot;

Timeout DotUpdateTimeout;
void DotUpdateCallback(void) {
    if (Dot.CurrentState == Incrementing && Dot.Value < DOT_MAX) {
        Dot.Value = (Dot.Value + DOT_FADE_STEP >= DOT_MAX) ? DOT_MAX : Dot.Value + DOT_FADE_STEP;
    } else if (Dot.CurrentState == Decrementing && Dot.Value > DOT_MIN) {
        Dot.Value = (Dot.Value - DOT_FADE_STEP <= DOT_MIN) ? DOT_MIN : Dot.Value - DOT_FADE_STEP;
    }
    Dot.Updated = true;

    if (Dot.Value != DOT_MAX && Dot.Value != DOT_MIN) {
        DotUpdateTimeout.attach_us(DotUpdateCallback, DOT_FADE_DURATION_US / REFRESH_RATE_US);
    } 
}

// Macro the tube callback function as each callback needs to be defined as
// a unique function (callbacks cannot have arguments and are not reentrant)
#define TUBE_CALLBACK(x) \
void Tube##x##UpdateCallback(void) { \
    int step = (DOT_MAX + (Tubes[x].FadeDuration / REFRESH_RATE_US) - 1) / (Tubes[x].FadeDuration / REFRESH_RATE_US); \
    bool activeTube = false; \
    for (int i = 0; i < 10; i++) { \
        Digit *digit = &Tubes[x].Digits[i]; \
        if (digit->CurrentState == Incrementing && digit->Value < DIGIT_MAX) { \
            digit->Value = (digit->Value + step >= DIGIT_MAX) ? DIGIT_MAX : digit->Value + step; \
            digit->Updated = true; \
        } else if (digit->CurrentState == Decrementing && digit->Value > DOT_MIN) { \
            digit->Value = (digit->Value - step <= DIGIT_MIN) ? DIGIT_MIN : digit->Value - step; \
            digit->Updated = true; \
        } \
        activeTube |= (digit->Value != DIGIT_MAX && digit->Value != DIGIT_MIN); \
    } \
    if (activeTube) { \
        Tube##x##UpdateTimeout.attach_us(Tube##x##UpdateCallback, DIGIT_FADE_DURATION_US / REFRESH_RATE_US); \
    } \
} 

Timeout Tube0UpdateTimeout;
TUBE_CALLBACK(0)
Timeout Tube1UpdateTimeout;
TUBE_CALLBACK(1)
Timeout Tube2UpdateTimeout;
TUBE_CALLBACK(2)
Timeout Tube3UpdateTimeout;
TUBE_CALLBACK(3)

void FadeInOutDigit(int T, int D, int Duration) {
    for (int i = 0; i < 10; i++) {
        Tubes[T].Digits[i].CurrentState = Decrementing;
    }
    if (D != -1) {
        Tubes[T].Digits[D].CurrentState = Incrementing;
    }
    Tubes[T].FadeDuration = Duration;

    if (T == 0) {
        Tube0UpdateTimeout.attach_us(Tube0UpdateCallback, DIGIT_FADE_DURATION_US / REFRESH_RATE_US);
    } else if (T == 1) {
        Tube1UpdateTimeout.attach_us(Tube1UpdateCallback, DIGIT_FADE_DURATION_US / REFRESH_RATE_US);
    } else if (T == 2) {
        Tube2UpdateTimeout.attach_us(Tube2UpdateCallback, DIGIT_FADE_DURATION_US / REFRESH_RATE_US);
    } else if (T == 3) {
        Tube3UpdateTimeout.attach_us(Tube3UpdateCallback, DIGIT_FADE_DURATION_US / REFRESH_RATE_US);
    } 
}

Ticker RefreshTicker, RngUpdateTicker;
int RngTickerIter;
void RngTickerUpdate(void) {


    if (--RngTickerIter == 0) {
        RngUpdateTicker.detach();
    }
}

// Callback from DS3231 interrupt (1Hz)
void RtcTickCallback(void) {
    RtcTick = true;
}

// Callback from RefreshTicker (REFRESH_RATE_US)
void RefreshTickCallback(void) {
    RefreshTick = true;
}

void RngTickCallback(void) {
    RngTick = true;
}

int main() {

    // Initialize pointers in global data structure
    for (int i = 0; i < NUM_TUBES; i++) {
        for (int j = 0; j < 10; j++) {
            Tubes[i].Digits[j].CurrentState = Decrementing;
            Tubes[i].Digits[j].Value = 0;
            Tubes[i].Digits[j].Updated = false;
        }
        Tubes[i].Refresh = false;
        Tubes[i].FadeDuration = 0;
    }

    Dot.CurrentState = Decrementing;
    Dot.Value = 0;
    Dot.Updated = false;

    RtcTick = false;
    RefreshTick = false;
    RngTick = false;

    int nextSecond, nextMinute, nextHour;
    int prevMinute, prevHour;
    bool startup = true;
    
    // Start I2C at 400kHz for DS3231 
    i2c.frequency(400000);
    
    // Start with HV PSU disabled
    HV_EnableOutput(false);
    
    TUSB322_Init();
    PCA9685_Init();
    DS3231_Init(RtcTickCallback);

    // Enable HV PSU
    HV_EnableOutput(true);
    
    // Set PCA9685 input voltage to highest possible
    PCA9685_SetVoltage(1.0);
    
    // swo.printf("CPU SystemCoreClock is %d Hz\r\n", SystemCoreClock);
    
    // Bump I2C frequency to 1MHz
    // i2c.frequency(1000000);

    // Setup a ticker to refresh the display at 1kHz
    RefreshTicker.attach_us(RefreshTickCallback, REFRESH_RATE_US);

    // Kick off the RNG ticker at 0.1Hz
    // Ticker rngTicker;
    // rngTicker.attach(RngTickCallback, DIGIT_RNG_REFRESH_RATE_S);

    // DS3231_SetTime(0, 55, 6, true);
    // DS3231_SetDate(0, 2, 12, 18, 0);

    while(1) {

        // Animate_Cycle_Basic();
        // Animate_Cycle_Analog();
        // Animate_Cycle_Low_Pwm();
        // Animate_Cycle_Pwm();
        // Animate_Cycle_Fade();
        // Animate_Cycle_Fade_Random();
        // Animate_Cycle_Fast();
        // Animate_Cycle_Fast_Random();

        // On every refresh tick, update the display with new values
        if (RefreshTick) {
            RefreshTick = false;

            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < 10; j++) {
                    if (Tubes[i].Digits[j].Updated) {
                        PCA9685_SetDigit(i, j, Tubes[i].Digits[j].Value);
                        Tubes[i].Digits[j].Updated = false;
                    }
                }
            }

            if (Dot.Updated) {
                PCA9685_SetDot(Dot.Value);
                Dot.Updated = false;
            }
        }

        // On every RTC tick, read and process the current time
        if (RtcTick) {
            RtcTick = false;

            // Save the previous time to check if digit(s) has changed
            prevMinute = nextMinute;
            prevHour = nextHour;

            DS3231_GetTime(&nextSecond, &nextMinute, &nextHour);

            // Update the display configuration based on the new/previous time
            if (startup || prevHour / 10 != nextHour / 10) {
                FadeInOutDigit(0, (nextHour / 10 != 0) ? nextHour / 10 : -1 , DIGIT_FADE_DURATION_US);
            } 
            if (startup || prevHour % 10 != nextHour % 10) {
                FadeInOutDigit(1, nextHour % 10, DIGIT_FADE_DURATION_US);
            } 
            if (startup || prevMinute / 10 != nextMinute / 10) {
                FadeInOutDigit(2, nextMinute / 10, DIGIT_FADE_DURATION_US);
            } 
            if (startup || prevMinute % 10 != nextMinute % 10) {
                FadeInOutDigit(3, nextMinute % 10, DIGIT_FADE_DURATION_US);
            }

            Dot.CurrentState = (Dot.CurrentState == Decrementing) ? Incrementing : Decrementing;
            DotUpdateTimeout.attach_us(DotUpdateCallback, DOT_FADE_DURATION_US / REFRESH_RATE_US);

            // Clear the startup run-once flag after the first iteration
            startup = false;
        }

        if (RngTick) {

            // Choose a random tube to refresh
            // Tube[rand() % 4].RefreshActive = true;

            // RngTickerIter = DIGIT_RNG_REFRESH_ITER * FADE_TICKS;
            // RngUpdateTicker.attach(RngTickerUpdate, (DIGIT_REFRESH_RATE_US/1000)/FADE_TICKS);
            // RngTick = false;
        }
    }
}

void I2C_Write(int DeviceAddress, char RegAddress, char *Data, int Length) {
    char buffer[I2C_MAX_BUFFER+1] = {0};

    if (Length > I2C_MAX_BUFFER) LED_Fault(1);

    buffer[0] = RegAddress;
    memcpy(&buffer[1], Data, Length);

    i2c.write(DeviceAddress << 1, buffer, Length + 1);
}

void I2C_Read(int DeviceAddress, char RegAddress, char *Data, int Length) {
    
    i2c.write(DeviceAddress << 1, &RegAddress, 1);
    i2c.read(DeviceAddress << 1, Data, Length);
}
